/*
 *
 * Difficulty (Update)
 *
 */

const events = require('events');
const utils = require('./utils');

////////////////////////////////////////////////////////////////////////////////

// Main Difficulty Function
const Difficulty = function(port, difficulty) {

  const _this = this;
  _this.options = difficulty;

  let lastTs, optimalDiff;
  let shareTime, sampleSize;
  const bufferSize = 15;

  // Cleanup diff
  this.disconnect = function(client) {
      lastTs = 0;
      optimalDiff = 0;
      shareTime = 0;
      sampleSize = 0;
  }

  // Update Difficulty on Share Submission
  this.updateDifficulty = function(client) {
    const ts = (Date.now() / 1000) | 0;

    if (!lastTs && optimalDiff != _this.options.initial) {
        lastTs = ts;
        sampleSize = shareTime = 0.000;
        optimalDiff = _this.options.initial;
        return;
    }

    // Bail-out early if we're still waiting on difficulty update
    if(optimalDiff != client.difficulty) {
        lastTs = ts;
        return;
    }

    shareTime += (ts - lastTs);
    sampleSize += 1;

    lastTs = ts;
    // Not enough sample to know, keep sampling
    if(sampleSize < bufferSize){
        return;
    }

    // We're living the dream
    // Try to keep the buffer in a sane state so that we keep living the dream
    if(sampleSize >= 1000) {
        sampleSize = shareTime = 0.000;
        return;
    }

    const coef = _this.options.targetTime / (shareTime / sampleSize);
    // Bail-out if we're good
    if((coef < (1 + _this.options.variance)) && (coef > (1 - _this.options.variance))) {
        return;
    }

    // We're out of "the zone", let's try to fix that
    // That "should"(tm) be the right new difficulty
    optimalDiff = utils.toFixed(client.difficulty * coef, 8);

    // Ensure we don't go higher than legal maximum and minimum
    // Also make sure we don't swing to hard after the initial correction
    if(optimalDiff > _this.options.maximum) {
        optimalDiff = _this.options.maximum;
    } else if(optimalDiff < _this.options.minimum) {
        optimalDiff = _this.options.minimum;
    } else if(optimalDiff > (client.difficulty * (1 + _this.options.variance)) && client.difficulty != _this.options.initial) {
        optimalDiff = utils.toFixed(client.difficulty * (1 + _this.options.variance), 8);
    } else if(optimalDiff < (client.difficulty * (1 - _this.options.variance)) && client.difficulty != _this.options.initial) {
        optimalDiff = utils.toFixed(client.difficulty * (1 - _this.options.variance), 8);
    }

    // Reset buffers then emit new optimal difficulty
    sampleSize = shareTime = 0.000;
    _this.emit('newDifficulty', client, optimalDiff);
  };

  // Manage Individual Clients
  this.manageClient = function(client) {
    const stratumPort = client.socket.localPort;
    if (stratumPort != port) {
      console.error('Handling a client which is not of this vardiff?');
    }
    client.on('submit', () => _this.updateDifficulty(client));
    client.on('socketDisconnect', () => _this.disconnect(client));
  };
};

module.exports = Difficulty;
Difficulty.prototype.__proto__ = events.EventEmitter.prototype;
